/*
 *  Utility to generate a script that allow the system to keep the current 
 *  configuration between system reboots. 
 *
 *  It is intended to be a help for the administrator who uses Winmasq as a 
 *  day to day administration and wishes to keep the current status of the i
 *  firewall when the system reboots.
 *
 *  Start Date: 28/10/97
 */      


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <linux/types.h>
#include <linux/timer.h>
#include <linux/socket.h>
#include <linux/sockios.h>
#include <linux/in.h>
#include <linux/inet.h>
#include <linux/netdevice.h> 
#include <linux/ip.h>
#include <linux/icmp.h>
#include <linux/tcp.h>
#include <linux/udp.h>
#include <linux/skbuff.h>
#include <linux/firewall.h>
#include <linux/ip_fw.h>

#define UNISTD 1
#include "masq.h"

#define FW_IN	1
#define FW_OUT	2
#define FW_FWD	4 
#define FW_MASQ 7

/*
 *   Opens the /proc/net/ip_* to gather all information 
 */

FILE *open_file(__u16 firewall)
{
    FILE *fitx;    

    switch (firewall)
        {
        case IP_FW_IN:   if((fitx=fopen(IP_FW_IN_FILE,"r"))==NULL) terminate(FILE_ERR);
                         break;
        case IP_FW_OUT:  if((fitx=fopen(IP_FW_OUT_FILE,"r"))==NULL) terminate(FILE_ERR);
                         break;
        case IP_FW_FWD:  if((fitx=fopen(IP_FW_FWD_FILE,"r"))==NULL) terminate(FILE_ERR);
                         break;
        default: terminate (ERROR);   /* Never reached */
        }
    return (fitx);
}

/*
 *       Proceses info lines from /proc/net files and convert data to
 *          struct ip_fw format to be re-processed 
 */

struct ip_fw *get_rule(FILE *fp, struct ip_fw *rule)
{
        int i,num;

        num=fscanf(fp, "%lX/%lX->%lX/%lX %16s %lX %hX %hu %hu %lu %lu",
                &(rule->fw_src.s_addr), &(rule->fw_smsk.s_addr),
                &(rule->fw_dst.s_addr), &(rule->fw_dmsk.s_addr),
                rule->fw_vianame, &(rule->fw_via.s_addr),
                &rule->fw_flg, &rule->fw_nsp, &rule->fw_ndp,
                &rule->fw_pcnt, &rule->fw_bcnt);
        for (i = 0; i < IP_FW_MAX_PORTS; i++)
        num+=fscanf(fp, "%hu", &rule->fw_pts[i]);
        num+=fscanf(fp, " A%hX X%hX", &rule->fw_tosand, &rule->fw_tosxor);
        if (num != RULE_ELEMENTS) return NULL;
        if (!strcmp("-", rule->fw_vianame))
                (rule->fw_vianame)[0] = '\0';
        rule->fw_src.s_addr=htonl(rule->fw_src.s_addr);
        rule->fw_smsk.s_addr=htonl(rule->fw_smsk.s_addr);
        rule->fw_dst.s_addr=htonl(rule->fw_dst.s_addr);
        rule->fw_dmsk.s_addr=htonl(rule->fw_dmsk.s_addr);
        rule->fw_next=NULL;
	return rule;
}


/*
 *
 *  Read rules from /proc/net files and show it to the user in a readable form.
 */   

void savefw(unsigned short firewall,FILE *script)
{
	int i,np;
	char fw;
	FILE *fitx;
	char *stract,*orig,*origmsk,*desti,*destimsk,*strprot, *sportstr, *dportstr, *strnum;
	char *line,c,*bidir;
	struct ip_fw *rule; 
	__u16 sprt,dprt,action,num=0; 

            
	fitx=open_file(firewall);

	if ((rule=(struct ip_fw *)malloc(sizeof(struct ip_fw)))==NULL) terminate(MEM_ERR);
	if ((line=(char *)malloc(LONGLINE+1))==NULL) terminate(MEM_ERR);
	if ((stract=(char *)malloc(BUFSIZE))==NULL) terminate(MEM_ERR);
	if ((strprot=(char *)malloc(BUFSIZE))==NULL) terminate(MEM_ERR);
	if ((orig=(char *)malloc(BUFSIZE))==NULL) terminate(MEM_ERR);
	if ((origmsk=(char *)malloc(BUFSIZE))==NULL) terminate(MEM_ERR);
	if ((desti=(char *)malloc(BUFSIZE))==NULL) terminate(MEM_ERR);
	if ((destimsk=(char *)malloc(BUFSIZE))==NULL) terminate(MEM_ERR);
	if ((sportstr=(char *)malloc(BUFSIZE))==NULL) terminate(MEM_ERR);
	if ((dportstr=(char *)malloc(BUFSIZE))==NULL) terminate(MEM_ERR);
 	if ((bidir=(char *)malloc(5))==NULL) terminate(MEM_ERR);

	while (!(isdigit( (c=fgetc(fitx)) )));
	ungetc(c,fitx);
	fgets(line,LONGLINE,fitx); 
	sscanf(line,"%d",&action);
	fw=set_fw(firewall);
        set_action(stract,action);
        fprintf(script,"${BIN}/lmasq -%c policy %s\n", fw, stract);
	fprintf(script,"${BIN}/lmasq -%c reset\n", fw);
      
	while (ungetc(fgetc(fitx),fitx)!=EOF) {
		if (get_rule(fitx,rule)==NULL) break; 
		strcpy(orig,(char *)inet_ntoa(rule->fw_src)); 
		strcpy(origmsk,(char *)inet_ntoa(rule->fw_smsk)); 
		strcpy(desti,(char *)inet_ntoa(rule->fw_dst)); 
		strcpy(destimsk,(char *)inet_ntoa(rule->fw_dmsk)); 

		sportstr=list_ports(sportstr,rule,rule->fw_nsp,0,FALSE);
		if (strstr(sportstr,"ALL")!=NULL) strcpy(sportstr,"0\0");
		dportstr=list_ports(dportstr,rule,rule->fw_ndp,rule->fw_nsp,FALSE);
		if (strstr(dportstr,"ALL")!=NULL) strcpy(dportstr,"0\0");

		set_protocol(strprot,(rule->fw_flg)&IP_FW_F_KIND);
		if (strstr(strprot,"all")!=NULL) strcpy(strprot,"\0");

		set_action(stract,((rule->fw_flg)>>2)<<2);
		if ((((rule->fw_flg)>>2)<<2)&IP_FW_F_BIDIR) strcpy(bidir,"-b\0"); 
			else strcpy(bidir,"\0");
		fprintf(script,"${BIN}/lmasq -%c add %s %s %s %s %s %s %s %s %s\n", fw,strprot,bidir,orig,origmsk,sportstr,desti,destimsk,dportstr,stract);
	}
     	free (stract); free(strprot); free(bidir);
        free(orig); free(desti); free(line); free(rule);
        free (origmsk); free(destimsk); free(sportstr); free(dportstr);
	fclose(fitx);    
}

int main (int argc,char *argv[])
{

  int opt,firewall=0;
  FILE *script;

  setuid(0);
  if (geteuid()!=0) {
	printf("You need root privileges to execute this program\n");
	exit (0);
  }

           /* Parameter Management  */

	while ((opt=getopt(argc, argv, "ifoa"))!=-1) 
		switch (opt) {
			case 'i':    firewall+=FW_IN;  /* Input Firewall */
					break;
			case 'o':    firewall+=FW_OUT; /* Output Firewall */
					break;
			case 'f':    firewall+=FW_FWD; /* Forward Firewall */ 
					break;
			case 'a':   firewall=FW_IN+FW_OUT+FW_FWD; /* Everything */
		            break;
	      
			default: fprintf(stderr, "Unknown option %c ",opt);
					 exit (-1);
		}
	if (firewall==0) firewall=FW_FWD;
	script=fopen(SCRIPT_FILE,"w");
	fprintf(script,"#!/bin/"SHELL"\n\nBIN="BINDIR"\n\n");
	fprintf(script,"for n in /lib/modules/`uname -r`/ipv4/ip_masq_*.o; do modprobe $n; done;\n\n");
	if ((firewall&FW_MASQ)&FW_IN)  savefw(IP_FW_IN,script);
	if ((firewall&FW_MASQ)&FW_OUT) savefw(IP_FW_OUT,script);
	if ((firewall&FW_MASQ)&FW_FWD)  savefw(IP_FW_FWD,script);
	fclose(script);
	chmod(SCRIPT_FILE,S_IRUSR|S_IWUSR|S_IXUSR);

	exit(0);
}

