/*
 *    Tercera utilitat de projecte sobre Conversio d'adreces IP
 *    
 *    Tipus de Programa: CLIENT TEXTE
 *    Objectiu: Manteniment (afegir, esborrar, consultar) de "rules" del Firewall 
 *    Data inici: 27/5/96
 * 
 *    Fase2: Separaci modular CLIENT/SERVIDOR (4/7/96)
 *
 *    lmasq.c: Creating a new version of masq/masqd functionality to act 
 *		directly on the firewall box as a root
 *		It has same functionality as Phase 1 (admrule, setpolicy, 
 *		masqstat) but all in on program 
 *
 *    Included in first alpha Masqd distributon on 30/7/96
 *
 *    26/10/97: Added bidirectional feature support
 *    26/10/97: Added range ports feature support
 */      


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <linux/types.h>
#include <linux/timer.h>
#include <linux/socket.h>
#include <linux/sockios.h>
#include <linux/in.h>
#include <linux/inet.h>
#include <linux/netdevice.h> 
#include <linux/ip.h>
#include <linux/icmp.h>
#include <linux/tcp.h>
#include <linux/udp.h>
#include <linux/skbuff.h>
#include <linux/firewall.h>
#include <linux/ip_fw.h>

#include "masq.h"

/*
 *   Omple la rule amb els parametres entrats (adreces, ports, accio)
 */

/*
 *   CLIENT: Prepares RULE to be sent to the server, 
 *           from parameters entered by user
 */

struct ip_fw *add_rule(struct ip_fw *rule, char **argv, int npar)
{
        unsigned short port;
   
		rule->fw_next=NULL;

        if (isalpha(argv[npar][0])) rule->fw_flg=get_protocol(argv[npar++]);
		if (argv[npar][0]=='-') {
			if (toupper(argv[npar][1])=='B') 
				rule->fw_flg+=IP_FW_F_BIDIR;
			else terminate(PARAMS_ERR);
			npar++;
		}

        rule->fw_src.s_addr=inet_addr(argv[npar++]); 
        rule->fw_smsk.s_addr=inet_addr(argv[npar++]); 
	rule->fw_nsp=extract_ports(argv[npar++],rule,0);
	if (rule->fw_nsp>1) rule->fw_flg+=IP_FW_F_SRNG;
        rule->fw_dst.s_addr=inet_addr(argv[npar++]); 
        rule->fw_dmsk.s_addr=inet_addr(argv[npar++]); 
	rule->fw_ndp=extract_ports(argv[npar++],rule,rule->fw_nsp);
	if (rule->fw_ndp>1) rule->fw_flg+=IP_FW_F_DRNG;
        rule->fw_flg|=get_action(argv[npar++]);

        return(rule); 
}


/*
 *   SERVER: Opens /proc/net files to gather firewall information 
 */

FILE *open_file(__u16 firewall)
{
    FILE *fitx;    

    switch (firewall)
        {
        case IP_FW_IN:   if((fitx=fopen(IP_FW_IN_FILE,"r"))==NULL) terminate(FILE_ERR);
                         break;
        case IP_FW_OUT:  if((fitx=fopen(IP_FW_OUT_FILE,"r"))==NULL) terminate(FILE_ERR);
                         break;
        case IP_FW_FWD:  if((fitx=fopen(IP_FW_FWD_FILE,"r"))==NULL) terminate(FILE_ERR);
                         break;
        case IP_FW_MSQ:  if((fitx=fopen(IP_FW_MSQ_FILE,"r"))==NULL) return NULL;
                         break;
        default: terminate (ERROR);   /* No hauria d'arribar-hi mai */
        }
    return (fitx);
}

/*
 *       Proceses info lines from /proc/net files and convert data to
 *          struct ip_fw format to be re-processed 
 */

struct ip_fw *get_rule(FILE *fp, struct ip_fw *rule)
{
        int i,num;

        num=fscanf(fp, "%lX/%lX->%lX/%lX %16s %lX %hX %hu %hu %lu %lu",
                &(rule->fw_src.s_addr), &(rule->fw_smsk.s_addr),
                &(rule->fw_dst.s_addr), &(rule->fw_dmsk.s_addr),
                rule->fw_vianame, &(rule->fw_via.s_addr),
                &rule->fw_flg, &rule->fw_nsp, &rule->fw_ndp,
                &rule->fw_pcnt, &rule->fw_bcnt);
        for (i = 0; i < IP_FW_MAX_PORTS; i++)
        num+=fscanf(fp, "%hu", &rule->fw_pts[i]);
        num+=fscanf(fp, " A%hX X%hX", &rule->fw_tosand, &rule->fw_tosxor);
        if (num != RULE_ELEMENTS) return NULL;
        if (!strcmp("-", rule->fw_vianame))
                (rule->fw_vianame)[0] = '\0';
        rule->fw_src.s_addr=htonl(rule->fw_src.s_addr);
        rule->fw_smsk.s_addr=htonl(rule->fw_smsk.s_addr);
        rule->fw_dst.s_addr=htonl(rule->fw_dst.s_addr);
        rule->fw_dmsk.s_addr=htonl(rule->fw_dmsk.s_addr);
        rule->fw_next=NULL;
	return rule;
}


/*
 *
 *  Read rules from /proc/net files and show it to the user in a readable form.
 */   

void list_rules(unsigned short firewall)
{
	int i,np;
	FILE *fitx;
	char *stract,*orig,*desti,*strprot, *strfire, *sportstr, *dportstr, *strnum;
	char *line,c,bidir;
	struct ip_fw *rule; 
	__u16 sprt,dprt,action,num=0; 

            
	fitx=open_file(firewall);

	if ((rule=(struct ip_fw *)malloc(sizeof(struct ip_fw)))==NULL) terminate(MEM_ERR);
	if ((line=(char *)malloc(LONGLINE+1))==NULL) terminate(MEM_ERR);
	if ((stract=(char *)malloc(BUFSIZE))==NULL) terminate(MEM_ERR);
	if ((strfire=(char *)malloc(BUFSIZE))==NULL) terminate(MEM_ERR);
        if ((strprot=(char *)malloc(BUFSIZE))==NULL) terminate(MEM_ERR);
	if ((orig=(char *)malloc(BUFSIZE))==NULL) terminate(MEM_ERR);
	if ((desti=(char *)malloc(BUFSIZE))==NULL) terminate(MEM_ERR);
        if ((sportstr=(char *)malloc(BUFSIZE))==NULL) terminate(MEM_ERR);
        if ((dportstr=(char *)malloc(BUFSIZE))==NULL) terminate(MEM_ERR);
        if ((strnum=(char *)malloc(BUFSIZE))==NULL) terminate(MEM_ERR);

	while (!(isdigit( (c=fgetc(fitx)) )));
	ungetc(c,fitx);
	fgets(line,LONGLINE,fitx); 
	sscanf(line,"%d",&action);
     	set_firewall(strfire,firewall);
        set_action(stract,action);
        printf("\nThe default action for %s firewall is %s\n",
                  strfire, stract);
        printf("Exceptions rule list\n");
	for(i=0;i<80;i++) printf("-"); printf("\n");
	printf("N   from          port             to            port     protocol bidir action\n");
	for(i=0;i<80;i++) printf("="); printf("\n");
      
	while (ungetc(fgetc(fitx),fitx)!=EOF) {
		if (get_rule(fitx,rule)==NULL) break; 
		strcpy(orig,(char *)inet_ntoa(rule->fw_src)); 
		strcpy(desti,(char *)inet_ntoa(rule->fw_dst)); 
		for (i=15-strlen(orig);i>0; i--) strcat(orig," \0");
                for (i=15-strlen(desti);i>0; i--) strcat(desti," \0");
		num++; 
		np=0;
		sportstr=list_ports(sportstr,rule,rule->fw_nsp,0,TRUE,TRUE);
		dportstr=list_ports(dportstr,rule,rule->fw_ndp,rule->fw_nsp,TRUE,FALSE);
                sprintf(strnum,"%d:",num);
                for (i=3-strlen(strnum);i>0; i--) strcat(strnum," \0");
                set_protocol(strprot,(rule->fw_flg)&IP_FW_F_KIND);
		for (i=0;i<strlen(strprot); i++) strprot[i]=toupper(strprot[i]);
                set_action(stract,((rule->fw_flg)>>2)<<2);
		if ((((rule->fw_flg)>>2)<<2)&IP_FW_F_BIDIR) bidir='Y'; 
			else bidir='N';
                for (i=4-strlen(strprot);i>0; i--) strcat(strprot," \0");
                for (i=7-strlen(stract);i>0; i--) strcat(stract," \0");		
		printf("%s%s  %s %s  %s  %s  %c  %s\n",
                        strnum,orig,sportstr,desti,dportstr,strprot,bidir,stract);
	}
     	free (stract); free(strprot); free(strfire);
        free(orig); free(desti); free(line); free(rule);
        free(strnum); free(sportstr); free(dportstr);
	fclose(fitx);    
}

/*
 *      Deletes a rule from firewall
 *
 */

struct ip_fw *del_rule(struct ip_fw *rule, char **argv, int npar, __u16 firewall)
{
     __u16 num,i;
     FILE *fitx;
     char *line;
 
     num=atoi(argv[npar++]);
     fitx=open_file(firewall);
     if ((line=(char *)malloc(LONGLINE+1))==NULL) terminate(MEM_ERR);
     fgets(line,LONGLINE,fitx);
     free(line);

     for(i=0;i<num;i++)
          if (get_rule(fitx,rule)==NULL) return NULL; 
    
     fclose(fitx);
     rule->fw_next=NULL;
     rule->fw_viadev=NULL;
     rule->fw_via.s_addr=0;
     for(i=0;i<IFNAMSIZ;i++) rule->fw_vianame[i]=0;
     return(rule);
}

void list_masqinf ()
{
	int i,j;
	FILE *f;
	char *line, c,*protocol, *saddrstr, *daddrstr, *sportstr, *mportstr;
	int linelong;
	unsigned int srcaddr, srcport, dstaddr, dstport, msqport; 

	if ((f=fopen(IP_FW_MSQ_FILE,"r"))==NULL) terminate(FILE_ERR);
	if ((line=(char *)malloc(LONGLINE+1))==NULL) terminate(MEM_ERR); 
	fgets(line,LONGLINE,f);
	printf("Protocol  From Address      From Port <-> Masqued Port  To Address    To Port\n\0");
	for(i=0;i<80;i++) printf("=");
	if ((protocol=(char *)malloc(BUFSIZE))==NULL) terminate(MEM_ERR); 
	if ((saddrstr=(char *)malloc(BUFSIZE))==NULL) terminate(MEM_ERR); 
	if ((daddrstr=(char *)malloc(BUFSIZE))==NULL) terminate(MEM_ERR); 
	if ((daddrstr=(char *)malloc(BUFSIZE))==NULL) terminate(MEM_ERR); 
	if ((sportstr=(char *)malloc(BUFSIZE))==NULL) terminate(MEM_ERR); 
	if ((mportstr=(char *)malloc(BUFSIZE))==NULL) terminate(MEM_ERR); 
        
	while (ungetc(fgetc(f),f)!=EOF) {
		i=-1; 
		fgets(line,LONGLINE,f);
		while(line[++i]==' ');
		if (isalpha(line[i]))  {
			sscanf(line,"%3s %8x%1c%4x %8x%1c%4x %4x",protocol,&srcaddr,&c,&srcport,
				&dstaddr,&c,&dstport,&msqport);
			srcaddr=htonl(srcaddr);
			strcpy(saddrstr,(char *)inet_ntoa(*((struct in_addr *)&srcaddr)));
			for (j=16-strlen(saddrstr);j>0; j--) strcat(saddrstr," \0");
			dstaddr=htonl(dstaddr);
			strcpy(daddrstr,(char *)inet_ntoa(*((struct in_addr *)&dstaddr)));	
			for (j=16-strlen(daddrstr);j>0; j--) strcat(daddrstr," \0");
			for (j=4-strlen(protocol);j>0; j--) strcat(protocol," \0");
			sprintf(sportstr,"%d",srcport);
			for (j=5-strlen(sportstr);j>0; j--) strcat(sportstr," \0");
			sprintf(mportstr,"%d",msqport);
			for (j=5-strlen(mportstr);j>0; j--) strcat(mportstr," \0");
			printf("  %s    %s      %s  <->    %s      %s  %d\n\0",
				protocol, saddrstr, sportstr, mportstr, 
 				daddrstr, dstport);
		}
	}

	free(saddrstr); free(sportstr);
	free(daddrstr); free(mportstr);
	free(protocol); 
	free(line); 
	fclose(f); 
}

/*
 *   Funcio de punt d'entrada al kernel.
 *   Executa l'acci del programa, en funcio dels parametres.
 */

/*
 *           Interacts with kernel through setsockopts call
 */ 

void ip_fw_opt(int command, int firewall, int policy, struct ip_fw *rule)
{
	int s,err;  
	void *optval;
	int domain, type, protocol;         
	int optlen,level,optname;

                 /* Obrim un socket amb aquests parametres */ 

	domain=AF_INET;
	type=SOCK_DGRAM;      /* Podria ser SOCK_RAW o SOCK_STREAM */
	protocol=IPPROTO_UDP; /* tamb IPPROTO_RAW o IPPROTO_TCP resp type */

	if ((s=socket(domain, type, protocol))<0) terminate(SOCKET_ERR); 

	level=SOL_IP;                                  /* Ataquem el nivell IP */
	optname=command | (firewall << IP_FW_SHIFT);   /* Comanda (XXX) */
	if (command!=IP_FW_POLICY) {
		optval=(void *)rule;                           /* Parametre de la comanda (Valor) */
		optlen=sizeof(*rule);                          /* Tamany del parametre */
	} else {
		optval=(void *)&policy;
		optlen=sizeof(policy);
	}
	if ((err=setsockopt(s, level, optname, optval, optlen))<0)
		terminate(SOCKOPT_ERR);
   
	close(s);
}
			
/*
 * 	MAIN function
 *           Parses command line and coordinates everything
 */

void do_initialize(struct ip_fw *rule)
{
	int policy=IP_FW_F_ACCEPT;
	
	ip_fw_opt (IP_FW_POLICY,IP_FW_IN,policy,NULL);
	ip_fw_opt (IP_FW_POLICY,IP_FW_OUT,policy,NULL);
	ip_fw_opt (IP_FW_POLICY,IP_FW_FWD,policy,NULL);
	ip_fw_opt (IP_FW_FLUSH,IP_FW_IN,0,rule); 
	ip_fw_opt (IP_FW_FLUSH,IP_FW_OUT,0,rule); 
	ip_fw_opt (IP_FW_FLUSH,IP_FW_FWD,0,rule);  
}


int main (int argc,char *argv[])
{
  struct ip_fw *rule;  
  int i,err,firewall, command, npar, policy;
  void *optval;
  char *param2;
  __u32 saddr, daddr;
  __u16 sport, dport, action; 

  setuid(0);
  if (geteuid()!=0) {
	printf("You need root privileges to execute this program\n");
	exit (0);
  }

           /* Gestio de parametres  */

  if (argc>NUM_PARAMS+1 || argc<2) terminate(PARAMS_ERR);
  if (argv[1][0]=='-') {

		switch (toupper(argv[1][1])) {
			case 'I':    firewall=IP_FW_IN;  /* Firewall d'entrada */
					break;
			case 'O':    firewall=IP_FW_OUT; /* Firewall de sortida */
					break;
			case 'F':    firewall=IP_FW_FWD; /* Firewall de encaminament */ 
					break;
			case 'M':   firewall=IP_FW_MSQ; /* Taula de masquerading */
		                        break;
	      
			default: terminate(PARAMS_ERR);
		};
		npar=2;
	}
	else { 
		npar=1;
		firewall=IP_FW_FWD; /* Firewall de encaminament per defecte*/ 
	}

        command=get_command(argv[npar],strlen(argv[npar]));
        if (command==GETLIST && firewall==IP_FW_MSQ) command=GETMSQINF;

	npar++;
	if ((rule=(struct ip_fw *)malloc(sizeof(struct ip_fw)))==NULL) terminate (MEM_ERR);
	switch (command) {
		case ADDRULE:	if ((argc-npar)<N_PAR_ADD) terminate(PARAMS_ERR);
				rule=add_rule(rule,argv,npar); 
				ip_fw_opt(IP_FW_APPEND, firewall, policy, rule);
				break;

		case DELRULE:	if ((argc-npar)!=N_PAR_DEL) terminate(PARAMS_ERR);
				rule=del_rule(rule,argv,npar,firewall); 
				ip_fw_opt(IP_FW_DELETE, firewall, policy, rule);
				break;

		case GETLIST:	list_rules(firewall);
				break;

		case RESET:	ip_fw_opt(IP_FW_FLUSH, firewall, policy, rule);
				break;

		case SETPOLICY: policy=get_action(argv[npar++]); 
				ip_fw_opt(IP_FW_POLICY, firewall, policy, rule);
				break;

		case GETMSQINF: list_masqinf();
				break;

		case INITIALIZE: do_initialize(rule);
				 break;
				
		default:  terminate(ERROR);  /* No hi hauria d'arribar mai! */
	};  
	free (rule); 
	exit(0);
}

