/*
 *    Tercera utilitat de projecte sobre Conversio d'adreces IP
 *    
 *    Tipus de Programa: CLIENT TEXTE
 *    Objectiu: Manteniment (afegir, esborrar, consultar) de "rules" del Firewall 
 *    Data inici: 27/5/96
 * 
 *    Fase2: Separaci modular CLIENT/SERVIDOR (4/7/96)
 *
 *    masq.c: Inici de conversio a modul CLIENT (9/7/96)
 *            Translating program text from catalan to english (28/7/96)
 *
 *    Ready for first alpha distribution (30/7/96)
 *
 *    30/10/97: Added bidirectional feature support
 *    30/10/97: Added range ports feature support
 */      


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <netdb.h>

#include <linux/types.h>
#include <linux/timer.h>
#include <linux/socket.h>
#include <linux/sockios.h>
#include <linux/in.h>
#include <linux/inet.h>
#include <linux/netdevice.h> 
#include <linux/ip.h>
#include <linux/icmp.h>
#include <linux/tcp.h>
#include <linux/udp.h>
#include <linux/skbuff.h>
#include <linux/firewall.h>
#include <linux/ip_fw.h>

#include "masq.h"

__u8 rcvbuf[CLN_RCVBUFSIZE];	 	/* Send buffer */
__u8 sndbuf[CLN_SNDBUFSIZE];	 	/* Receive buffer */


/*
 *   Omple la rule amb els parametres entrats (adreces, ports, accio)
 */

/*
 *   CLIENT: Prepares RULE to be sent to the server, 
 *           from parameters entered by user
 */

struct ip_fw *make_rule(struct ip_fw *rule, char **argv, int npar)
{
        unsigned short port;
   
	rule->fw_next=NULL;
        if (isalpha(argv[npar][0])) rule->fw_flg=get_protocol(argv[npar++]);
               if (argv[npar][0]=='-') {
                       if (toupper(argv[npar][1])=='B')
                               rule->fw_flg+=IP_FW_F_BIDIR;
                       else terminate(PARAMS_ERR);
                       npar++;
               }

        rule->fw_src.s_addr=inet_addr(argv[npar++]); 
        rule->fw_smsk.s_addr=inet_addr(argv[npar++]); 
        rule->fw_nsp=extract_ports(argv[npar++],rule,0);
	if (rule->fw_nsp>1) rule->fw_flg+=IP_FW_F_SRNG;
        rule->fw_dst.s_addr=inet_addr(argv[npar++]); 
        rule->fw_dmsk.s_addr=inet_addr(argv[npar++]); 
        rule->fw_ndp=extract_ports(argv[npar++],rule,rule->fw_nsp);
	if (rule->fw_ndp>1) rule->fw_flg+=IP_FW_F_DRNG;
        rule->fw_flg|=get_action(argv[npar++]);

        return(rule); 
}

/*
 *      Compares two operations.
 */

int checkadd (struct masqmsg *sndmsg, struct masqmsg *rcvmsg)
{
	struct ip_fw *sndrul, *rcvrul;
	int ports,i;
	
	sndrul=sndmsg->info.rule;
	rcvrul=rcvmsg->info.rule;
	ports=TRUE;
	for (i=0;i<IP_FW_MAX_PORTS;i++) 
		if (sndrul->fw_pts[i]!=rcvrul->fw_pts[i]) {
			ports=FALSE;
			break;
		}
 
	if (
		sndmsg->firewall==rcvmsg->firewall &&
		sndmsg->numrule==rcvmsg->numrule &&
		sndrul->fw_src.s_addr==rcvrul->fw_src.s_addr &&
		sndrul->fw_dst.s_addr==rcvrul->fw_dst.s_addr &&
		sndrul->fw_smsk.s_addr==rcvrul->fw_smsk.s_addr &&
		sndrul->fw_dmsk.s_addr==rcvrul->fw_dmsk.s_addr &&
		sndrul->fw_flg==rcvrul->fw_flg &&
		sndrul->fw_nsp==rcvrul->fw_nsp &&
		sndrul->fw_ndp==rcvrul->fw_ndp &&
		ports) return TRUE; 
			else return FALSE;
}

int checkdel (struct masqmsg *sndmsg, struct masqmsg *rcvmsg)
{
	struct ip_fw *rcvrul;
	char *strprot, *addrstr1, *addrstr2, ok;
	int port;

	rcvrul=rcvmsg->info.rule;
	
	if ( sndmsg->firewall!=rcvmsg->firewall || 
		sndmsg->numrule!=rcvmsg->numrule) 
			return 0;

      	if ((strprot=(char *)malloc(30))==NULL || 
		(addrstr1=(char *)malloc(16))==NULL || 
		(addrstr2=(char *)malloc(16))==NULL)
			terminate (MEM_ERR);
	strcpy(addrstr1,(char *)inet_ntoa(*((struct in_addr *)(&(rcvrul->fw_src)))));
	strcpy(addrstr2,(char *)inet_ntoa(*((struct in_addr *)(&(rcvrul->fw_smsk)))));
	if (rcvrul->fw_nsp) port=rcvrul->fw_pts[0];
		else port=0; 
	printf("\n *********  W A R N I N G *********\n");
	printf("Are you sure you want to delete this rule?\n");
	printf("------------------------------------------\n");
	printf("Protocol: %s\n",set_protocol(strprot,(rcvrul->fw_flg)&IP_FW_F_KIND));
	printf("From: %s netmask: %s port: %d\n", addrstr1, addrstr2, port);

	strcpy(addrstr1,(char *)inet_ntoa(*((struct in_addr *)(&(rcvrul->fw_dst)))));
	strcpy(addrstr2,(char *)inet_ntoa(*((struct in_addr *)(&(rcvrul->fw_dmsk)))));
	if (rcvrul->fw_nsp && rcvrul->fw_ndp) port=rcvrul->fw_pts[1];
		else if (rcvrul->fw_ndp) port=rcvrul->fw_pts[0]; 
			else port=0; 
	printf("To: %s netmask: %s port: %d\n", addrstr1, addrstr2, port);
	printf("Action to perform: %s\n",set_action(strprot,((rcvrul->fw_flg)>>2)<<2));
	printf("\n Is this OK? (y/n) ");
	free (strprot); free(addrstr1); free(addrstr2);
	while (isspace(ok=tolower(fgetc(stdin))));
	if (ok=='y') return TRUE;
		else return FALSE;
}

int checkop (struct masqmsg *sndmsg, struct masqmsg *rcvmsg)
{
	switch(sndmsg->command)	{
		case ADDRULE:	
			return (checkadd(sndmsg, rcvmsg));
		case DELRULE:
			return (checkdel(sndmsg, rcvmsg));
	}
}

/*
 *  Creates a socket and connect to the server
 */

int open_connection(int *sk, struct hostent *host)
{
	struct sockaddr_in ser_addr;
	int ser_addr_len,nrec;
	void *optval;	
	char addr[6];
	struct masqmsg *msg;

	if ( (*sk = socket (AF_INET, SOCK_STREAM, IPPROTO_TCP))<0)
		terminate (SOCKET_ERR);

	datacpy(addr,host->h_addr_list[0],4);
	ser_addr.sin_family = AF_INET;
	ser_addr.sin_port=htons(SRVPORT);
	ser_addr.sin_addr.s_addr=(unsigned char)(addr[3])<<24|
               (unsigned char)(addr[2])<<16|(unsigned char)(addr[1])<<8|
               (unsigned char)(addr[0]);
	ser_addr_len = sizeof (ser_addr);

	if ( connect(*sk, (struct sockaddr *)&ser_addr, ser_addr_len)<0 )
		return 0;

	nrec=recv(*sk, &rcvbuf, CLN_RCVBUFSIZE, 0);
	if (nrec>=MINPKTSIZE) msg=(struct masqmsg *)&rcvbuf;
		else return 0; 
	if (msg->command==IDENTIFY) return 1;
		else if (msg->command==REJECT) {
			printf("Computer not allowed to connect to server %s\n",
				host->h_name);
			terminate (SILENT);
		} else  return 0;
}

/*
 *  Creates a message to be sent 
 */

struct masqmsg *make_msg(int command, int firewall, int param, struct ip_fw *rule)
{
	struct masqmsg *msg;
	
	msg=(struct masqmsg *)malloc(sizeof(struct masqmsg));
	msg->command=command;
	msg->firewall=firewall;
	
	if (rule) msg->info.rule=rule; 
		else msg->info.rule=NULL;

	msg->numrule=0; 
	msg->policy=0;
	switch (command) {
	
		case IDENTITY:	msg->numrule=1; break;
		case ADDRULE:	msg->numrule=1; break;
		case DELRULE:	msg->numrule=param; break;
		case SETPOLICY:	msg->policy=param; break;
	}

	return (msg);

}


/* 
 *  Sends a message to the server
 *  Deletes msg (frees it)
 */

void send_msqmsg(int sk, struct masqmsg *msg, int msgtofree)
{
	int i=0;
	__u32 crc;
	sndbuf[i++]=msg->command;
	sndbuf[i++]=msg->firewall;
	sndbuf[i++]=msg->policy>>8;
	sndbuf[i]=msg->policy&0x0F;
	i+=2;
	sndbuf[i++]=msg->flag;
	sndbuf[i++]=msg->numrule;
	if (msg->command==IDENTITY) 
		i+=datacpy(&sndbuf[i],msg->info.raw,strlen(msg->info.raw));
	if (msg->numrule && msg->command!=DELRULE && msg->command!=IDENTITY) 
		i+=datacpy(&sndbuf[i],msg->info.rule,sizeof(struct ip_fw));

	sndbuf[4]=i+sizeof(__u32);
	crc=checkcrc(&sndbuf,i);
	i+=datacpy(&sndbuf[i],&crc,sizeof(__u32));
	
	send(sk,&sndbuf, i, 0);
	if (msgtofree && msg->numrule && msg->command!=DELRULE) free(msg->info.raw); 
	if (msgtofree) free (msg);
}		

/*
 *   Send a confirm message (CLNACK)
 */

inline void send_ack(int sk)
{
        send_msqmsg(sk,make_msg(CLNACK,0,0,NULL),1);
}

/*
 *   Send a Non conform message (CLNNAK)
 */

inline void send_nak(int sk)
{
        send_msqmsg(sk,make_msg(CLNNAK,0,0,NULL),1);
}



/*
 *   Receives a complex message from the server
 *   msg MUST be allocated!
 */

int recv_msqmsg(int sk, struct masqmsg *msg)
{
        int nrec,num=0;
        char *info;

        nrec=recv(sk, &rcvbuf, CLN_RCVBUFSIZE, 0);
        if (nrec<MINPKTSIZE && nrec>0) terminate(PROTO_ERR);
                else if(nrec<0) return 0;

        msg->command=rcvbuf[0];
        msg->firewall=rcvbuf[1];
	msg->policy=rcvbuf[2]<<8|rcvbuf[3];
        msg->longpkt=rcvbuf[4];
        msg->flag=rcvbuf[5];
        msg->numrule=rcvbuf[6];
        if (msg->numrule) {
                info=(char *)malloc(CLN_RCVBUFSIZE);
                num=datacpy(info,&rcvbuf[7],msg->longpkt-MINPKTSIZE);
        }
        else
                info=NULL;
        msg->info.raw=info;
 	datacpy(&(msg->crc),&rcvbuf[7+num],sizeof(__u32));
        if (msg->crc==checkcrc(rcvbuf,7+num)) return 1;
		else return 0;
}


/*
 *   Request user identification and sends it to the server
 */

int identify_to_server(int sk)
{
	int nrec,i;
	char *user, *password,*pwcryp, key[2];
	struct masqmsg *rcvmsg, *sndmsg;

	if ((user=(char *)calloc(BUFSIZE,1))==NULL)  
		terminate(MEM_ERR);
	do {
		printf("Enter username: ");
		fgets(user,BUFSIZE,stdin);
		password=(char *)getpass("Password: ");
        	for (i=0; i<strlen(user);i++)
                	user[i]=tolower(user[i]);
        	user[i]='\0';
		key[0]=user[0]*3;
        	key[1]=user[0]*user[0];
		pwcryp=(char *)crypt(password,key);
		sndmsg=make_msg(IDENTITY,0,0,NULL);
		(sndmsg->info.raw)=(char *)calloc(strlen(user)+strlen(pwcryp)+1,1);
		if (sndmsg->info.raw==NULL)
			{ free (user); terminate(MEM_ERR); }
		strcpy(sndmsg->info.raw,user);
		strcat(sndmsg->info.raw,pwcryp);
		send_msqmsg(sk,sndmsg,1);
		nrec=recv(sk, &rcvbuf, CLN_RCVBUFSIZE, 0);
		if (nrec>=MINPKTSIZE) rcvmsg=(struct masqmsg *)&rcvbuf;
			else {free (user); terminate(PROTO_ERR); } 
	}
	while (!(rcvmsg->command==ACCEPT) && !(rcvmsg->command==REJECT)); 
	free (user);
	if (rcvmsg->command==ACCEPT) return 1;
		else return 0; 
} 

void talkone (int sk, struct masqmsg *sndmsg) 
{
	int nrec;
	char confirm;
	struct masqmsg *rcvmsg;
	
	if (sndmsg->command==ADDRULE || sndmsg->command==DELRULE) confirm=TRUE;
		else confirm=FALSE;
	send_msqmsg(sk,sndmsg,0);
	
	if (confirm) {
		rcvmsg=(struct masqmsg *)malloc(sizeof(struct masqmsg));
		recv_msqmsg(sk, rcvmsg);
		if (!(rcvmsg->command==CONFIRMOP)) {
			if (sndmsg->command==ADDRULE && rcvmsg->command==SRVNAK) { 
				printf("Unable to ADD specified rule\n");
				terminate(SILENT);
				}
			if (sndmsg->command==DELRULE && rcvmsg->command==SRVNAK) {
				printf("Unable to DELETE rule number %d\nDoes it really exist?\n",sndmsg->numrule); 
				terminate (SILENT);
				}
			terminate (PROTO_ERR);
		}
	
		if (checkop(sndmsg,rcvmsg)) send_ack(sk); 
			else {
				send_nak(sk);	
				free (rcvmsg);				
				if(sndmsg->command==ADDRULE) terminate(PROTO_ERR);
					else {
						printf("Operation NOT performed\n");
						terminate (SILENT);
					}
			}
		free (rcvmsg);				
	}
	free (sndmsg);
	nrec=recv(sk, &rcvbuf, CLN_RCVBUFSIZE, 0);
	if (nrec>=MINPKTSIZE) rcvmsg=(struct masqmsg *)&rcvbuf;
		else terminate(PROTO_ERR); 
	if (!(rcvmsg->command==SRVACK)) terminate(PROTO_ERR); 
		
}

/*
 * Closes the connection with the server;
 */

void close_connection(int sk)
{
	int nrec;
	struct masqmsg *rcvmsg;

	send_msqmsg(sk,make_msg(DISCONNECT,0,0,NULL),1);
	nrec=recv(sk, &rcvbuf, CLN_RCVBUFSIZE, 0);
	if (nrec>=MINPKTSIZE) rcvmsg=(struct masqmsg *)&rcvbuf;
		else terminate(PROTO_ERR); 
	if (!(rcvmsg->command==SRVACK)) terminate(PROTO_ERR); 
	close(sk);
	printf("\nOperation succesfull!\n");
}

/*
 *  MIXED -> Has to be converted to CLIENT/SERVER form.
 *
 *  Read rules from /proc/net files and show it to the user in a readable form.
 */   

int list_rules(int sk, int firewall)
{
	int i,np,nrec,num=0;
	char *stract,*orig,*desti,*strprot, *strfire, *sportstr, *dportstr, *strnum;
	char bidir;
	struct ip_fw *rule;
	__u16 sprt,dprt; 
	struct masqmsg *msg;
	
	msg=(struct masqmsg *)calloc(1,sizeof(struct masqmsg));
	send_msqmsg(sk,make_msg(GETLIST,firewall,0,NULL),1);
	if (!recv_msqmsg(sk, msg)) terminate(PROTO_ERR);

      	if ((stract=(char *)calloc(BUFSIZE,1))==NULL) terminate(MEM_ERR);
      	if ((strfire=(char *)calloc(BUFSIZE,1))==NULL) terminate(MEM_ERR);
      	if ((strprot=(char *)calloc(BUFSIZE,1))==NULL) terminate(MEM_ERR);
	if ((orig=(char *)calloc(BUFSIZE,1))==NULL) terminate(MEM_ERR);
	if ((desti=(char *)calloc(BUFSIZE,1))==NULL) terminate(MEM_ERR);
	if ((sportstr=(char *)calloc(BUFSIZE,1))==NULL) terminate(MEM_ERR);
	if ((dportstr=(char *)calloc(BUFSIZE,1))==NULL) terminate(MEM_ERR);
	if ((strnum=(char *)calloc(BUFSIZE,1))==NULL) terminate(MEM_ERR);

	while (msg->command==SENDLIST) {
          	num++; 
		if (num==1) {
			set_firewall(strfire,msg->firewall);
			set_action(stract,msg->policy);
			printf("\nThe default action for %s firewall is %s\n",
				strfire, stract);
			printf("Exceptions rule list\n");   
			for(i=0;i<80;i++) printf("-"); printf("\n");
	   	        printf("N\xba   from          port             to            port     protocol bidir action\n");
			for(i=0;i<80;i++) printf("="); printf("\n");
		}
		if (!(msg->numrule)) {
			send_nak(sk);
			free (stract); free(strprot);
			free(orig); free(desti); free(msg); free (sportstr);
			free(dportstr); free(strnum);
			nrec=recv(sk, &rcvbuf, CLN_RCVBUFSIZE, 0);
			if (nrec>=MINPKTSIZE) msg=(struct masqmsg *)&rcvbuf;
				else terminate(PROTO_ERR); 
			if (!(msg->command==SRVACK)) terminate(PROTO_ERR); 
			return 1;
		}
		else {
			rule=msg->info.rule;
			send_ack(sk);
			strcpy(orig,(char *)inet_ntoa(rule->fw_src));
			strcpy(desti,(char *)inet_ntoa(rule->fw_dst));
                	for (i=15-strlen(orig);i>0; i--) strcat(orig," \0");
                	for (i=15-strlen(desti);i>0; i--) strcat(desti," \0");
			np=0;
	                sportstr=list_ports(sportstr,rule,rule->fw_nsp,0,TRUE,TRUE);
         	        dportstr=list_ports(dportstr,rule,rule->fw_ndp,rule->fw_nsp,TRUE,FALSE);
			sprintf(strnum,"%d:",num);
                	for (i=3-strlen(strnum);i>0; i--) strcat(strnum," \0");
			set_protocol(strprot,(rule->fw_flg)&IP_FW_F_KIND); 
			for (i=0;i<strlen(strprot); i++) strprot[i]=toupper(strprot[i]);
			set_action(stract,((rule->fw_flg)>>2)<<2); 
			if ((((rule->fw_flg)>>2)<<2)&IP_FW_F_BIDIR) bidir='Y';
                        else bidir='N';
                	for (i=4-strlen(strprot);i>0; i--) strcat(strprot," \0");
                	for (i=7-strlen(stract);i>0; i--) strcat(stract," \0");
                  printf("%s%s  %s %s  %s  %s  %c  %s\n",
                         strnum,orig,sportstr,desti,dportstr,strprot,bidir,stract);
			if (!recv_msqmsg(sk, msg)) terminate(PROTO_ERR);
		}
	}
	if (msg->command!=SRVACK) terminate(PROTO_ERR);
	free (stract); free(strprot); free(strfire);
	free(orig); free(desti); free(msg); 
	free(strnum); free(sportstr); free(dportstr);
	return 0;
}

void masq_inf(int sk)
{
	int i,np,nrec,num=0;
	char *stract,*saddrstr,*daddrstr,*strprot,*sportstr,*dportstr,*mportstr;
	struct masqinf *mqentry;
	struct masqmsg *msg;
	__u16 sprt,dprt; 
	
	if ((mqentry=(struct masqinf *)malloc(sizeof(struct masqinf)))==NULL ||
		(msg=(struct masqmsg *)malloc(sizeof(struct masqmsg)))==NULL ||
		(saddrstr=(char *)malloc(BUFSIZE))==NULL || 
		(daddrstr=(char *)malloc(BUFSIZE))==NULL ||
		(sportstr=(char *)malloc(BUFSIZE))==NULL ||
		(dportstr=(char *)malloc(BUFSIZE))==NULL ||
		(mportstr=(char *)malloc(BUFSIZE))==NULL)
			terminate (MEM_ERR);
	send_msqmsg(sk,make_msg(GETMSQINF,IP_FW_MSQ,0,NULL),1);
	if (!recv_msqmsg(sk, msg)) terminate(PROTO_ERR);

      	if ((strprot=(char *)malloc(BUFSIZE))==NULL) terminate(MEM_ERR);

	printf("Protocol  From Address      From Port <-> Masqued Port  To Address    To Port\n\0");     
	for(i=0;i<80;i++) printf("=");

	while (msg->command==SENDMSQINF) {
          	num++; 
		mqentry=msg->info.masq;
		strcpy(saddrstr,(char *)inet_ntoa(*((struct in_addr *)(&(mqentry->saddr)))));
		strcpy(daddrstr,(char *)inet_ntoa(*((struct in_addr *)(&(mqentry->daddr)))));
                for (i=15-strlen(saddrstr);i>0; i--) strcat(saddrstr," \0");
                for (i=15-strlen(daddrstr);i>0; i--) strcat(daddrstr," \0");
		set_protocol(strprot,ntohs(mqentry->protocol)); 
		for (i=0;i<strlen(strprot); i++) strprot[i]=toupper(strprot[i]);
		for (i=4-strlen(strprot); i>0; i--) strcat (strprot," \0");
		sprintf(sportstr,"%d",ntohs(mqentry->sport));
		sprintf(dportstr,"%d",ntohs(mqentry->dport));
		sprintf(mportstr,"%d",ntohs(mqentry->mport));
		for (i=5-strlen(sportstr); i>0; i--) strcat (sportstr," \0");
		for (i=5-strlen(dportstr); i>0; i--) strcat (dportstr," \0");
		for (i=5-strlen(mportstr); i>0; i--) strcat (mportstr," \0");
		printf("  %s    %s      %s  <->    %s      %s  %s\n\0",
			strprot, saddrstr, sportstr, mportstr,	
			daddrstr, dportstr);
		send_ack(sk);	
		if (!recv_msqmsg(sk, msg)) terminate(PROTO_ERR);
	}
	if (msg->command!=SRVACK) terminate (PROTO_ERR); 
	free (strprot); free(saddrstr); free(daddrstr);
	free (sportstr); free(dportstr); free(mportstr);
}

/*
 *  Manages the comunication with the server
 */

void startxmit (struct hostent *host, int command, int firewall, int param, struct ip_fw *rule)
{
	int sk;		
	struct masqmsg *msg;

	if (!open_connection(&sk, host)) terminate (HOST_ERR);
 	if (!identify_to_server(sk)) terminate (IDENT_ERR);
	switch (command) {
		
		case	ADDRULE: 	msg=make_msg(command, firewall, param, rule);
					talkone(sk,msg);
					list_rules(sk, firewall);	
					break;
		case	DELRULE:
		case	SETPOLICY:
		case	RESET:		msg=make_msg(command, firewall, param, NULL);
					talkone(sk,msg);
					list_rules(sk, firewall);	
					break;
		case	INITIALIZE: 	msg=make_msg(command, firewall, param, NULL);
					talkone(sk,msg);
					list_rules(sk, IP_FW_IN);	
					list_rules(sk, IP_FW_OUT);	
					list_rules(sk, IP_FW_FWD);	
					break;	
		case	GETLIST:	list_rules(sk, firewall);
					break;
		case 	GETMSQINF:	masq_inf(sk);
                                	break;
		case	GENSCRIPT:	msg=make_msg(command, firewall, param, NULL);
					talkone(sk,msg);
					break;
		default:  terminate(ERROR);  /* Theorically unreachable */
	}
	close_connection(sk);
}
			
/*
 *  CLIENT: Main function
 *           Parses command line and coordinates everything
 */


int main (int argc,char *argv[])
{
	struct ip_fw *rule;  
	int i,firewall, command, npar, param;
	struct hostent *host;

           /* Gestio de parametres  */

	if (argc>NUM_PARAMS+1 || argc<2) terminate(PARAMS_ERR);
	if (argv[1][0]=='-') { 
		switch (toupper(argv[1][1])) {
			case 'I':   firewall=IP_FW_IN;  /* Firewall d'entrada */
				break;
			case 'O':   firewall=IP_FW_OUT; /* Firewall de sortida */
				break;
 			case 'F':   firewall=IP_FW_FWD; /* Firewall de encaminament */ 
				break;
			case 'M':   firewall=IP_FW_MSQ; /* Taula de masquerading */ 
				break;
			default: terminate(PARAMS_ERR);
		};
		npar=2;
	}
	else { 
		npar=1;
		firewall=IP_FW_FWD; /* Firewall de encaminament per defecte*/ 
	}

	command=get_command(argv[npar],strlen(argv[npar]));
	if (command==GETLIST && firewall==IP_FW_MSQ) command=GETMSQINF;

	npar++;

	if ((rule=(struct ip_fw *)malloc(sizeof(struct ip_fw)))==NULL) terminate (MEM_ERR);
	switch (command) {
		case ADDRULE:  	if ((argc-npar)<N_PAR_ADD) terminate(PARAMS_ERR);
				rule=make_rule(rule,argv,npar); 
				break;

		case DELRULE:	if ((argc-npar)<N_PAR_DEL) terminate(PARAMS_ERR);
				param=atoi(argv[npar++]);
				break;

		case SETPOLICY:	if ((argc-npar)<N_PAR_POLICY) terminate(PARAMS_ERR);
				param=get_action(argv[npar++]);	

		case GETLIST:  
		case GETMSQINF:	
		case RESET:	
		case GENSCRIPT:
		case INITIALIZE: break;
 
		default:  terminate(ERROR);  /* Theorically unreachable */
	};  
	if (argv[argc-1][0]=='-') { 
		if ((host=gethostbyname(argv[argc-1]+1))==NULL) terminate(HOST_ERR);
	}
	else  
		if ((host=gethostbyname("localhost"))==NULL) terminate(HOST_ERR);
	startxmit(host, command, firewall, param, rule);
	free (rule); 
	exit(0);
}

