/*
 *  Definitions for masq.c masqd.c masqutil.c
 *
 */

/*****************************************************************
 *   USER DEFINABLE CONSTANTS
 *****************************************************************/ 

/* 
 *  This is the file and place where the user list and passwords are kept
 *  to allow masqd perform user authentification 
 */

#define USER_FILE	  "/etc/masqdusers\0"

/*
 * This is the number of retries you want the masqd server allow the client to  
 * guess its username and password. 
 */

#define LOGIN_COUNT	3	

/*
 * This is the TCP port masqd will listen to and the one masq will connect to.
 * in othe words, it is the port they use to comunicate.
 */

#define SRVPORT		9500

/*
 *  If you experience some problems to connect from Windows95 clients, 
 *  undefine this. This will disable the reverse resolve from IP address
 *  to hostname and disabling this security check
 */

#define	SECURE_IP 

/*
 *  Definitions for list2script utility
 */

/*
 *  Name of the destination script
 */
 
#define SCRIPT_FILE    "/etc/init.d/rc.masquerade"

/*
 *  Name of the shell to use in first line of script
 */

#define SHELL "bash"

/*
 *  Directori where 'lmasq' is placed
 */

#define BINDIR  "/usr/local/bin"

/*****************************************************************
 *   END OF USER DEFINABLE CONSTANTS
 *****************************************************************/ 

/*
 *   Constants generals
 */

#define	TRUE	0x1 
#define FALSE	0x0
#define ERROR	0xFF

/*
 *   Valors d'error
 */

#define SILENT		0
#define PARAMS_ERR	0xFE
#define SOCKET_ERR	0xFD
#define SOCKOPT_ERR     0xFC
#define MEM_ERR		0xFB
#define FILE_ERR        0xFA
#define CHILD_ERR	0xF9
#define HOST_ERR	0xF8
#define IDENT_ERR	0xF7
#define PROTO_ERR	0xF6
#define RANGE_ERR 	0xF5

/*
 *  Constants de fitxers
 */

#define IP_FW_IN_FILE     "/proc/net/ip_input\0"
#define IP_FW_OUT_FILE    "/proc/net/ip_output\0"
#define IP_FW_FWD_FILE    "/proc/net/ip_forward\0"
#define IP_FW_MSQ_FILE    "/proc/net/ip_masquerade\0"  
/* For debug purpose only */
/* #define IP_FW_MSQ_FILE    "/root/projecte/masqtest\0"  */


/*
 *   Altres
 */

#define NUM_PARAMS	12
#define LONGLINE        110
#define BUFSIZE		30 

#define N_PAR_ADD       7 
#define N_PAR_DEL       1
#define N_PAR_POLICY	1	
#define N_PAR_LIST      0         
#define N_PAR_RESET     0         
#define N_PAR_MSQINF    0         

#define IP_FW_LIST	74
#define IP_FW_MSQ	10
#define NUM_RETRY	3
#define CLN_RCVBUFSIZE	1024
#define CLN_SNDBUFSIZE	256
#define SRV_RCVBUFSIZE  CLN_SNDBUFSIZE
#define SRV_SNDBUFSIZE  CLN_RCVBUFSIZE
#define MINPKTSIZE	11	
#define USER_LONG	32
#define RULE_ELEMENTS   23


/*
 *   Communications protocol   
 */

#define IDENTIFY	0x81
#define IDENTITY	0x1
#define ACCEPT		0x82
#define REJECT		0x83
#define GETLIST		0x2
#define SENDLIST	0x84
#define ADDRULE         0x3
#define DELRULE		0x4
#define SETPOLICY	0x5
#define CONFIRMOP	0x85
#define RESET		0x6
#define GETMSQINF	0x7
#define SENDMSQINF	0x87
#define CLNACK		0x8
#define SRVACK		0x88
#define CLNNAK		0x9
#define SRVNAK		0x89
#define DISCONNECT	0xA
#define INITIALIZE	0xB
#define GENSCRIPT	0xC
	
#include <linux/types.h>

struct masqinf {
        __u16           protocol;       /* Which protocol are we talking? */
        __u16           sport, dport, mport;    /* src, dst & masq ports */
        __u32           saddr, daddr, maddr;    /* src, dst & masq addresses */
	__u32		initseq;
	__u16 		delta,pdelta;
	__u32		exptime;                   /* Expiration time */
};

struct masqmsg {
	__u8 command;
	__u8 firewall;
	__u16 policy;
	__u8 longpkt;
	__u8 flag;
	__u8 numrule;
	union  {
		struct ip_fw *rule;
		struct masqinf *masq;
		char *raw;
		} info;
	__u32 crc;
	};

/*
 *   Headers  
 */
#ifndef UNISTD
char *crypt(char *text, char* key);
#endif
