/*
 *    Tercera utilitat de projecte sobre Conversio d'adreces IP
 *    
 *    Tipus de Programa: CLIENT TEXTE
 *    Objectiu: Manteniment (afegir, esborrar, consultar) de "rules" del Firewall 
 *    Data inici: 27/5/96
 * 
 *    Fase2: Separaci modular CLIENT/SERVIDOR (4/7/96)
 *
 *    masqd.c:  Inici de conversio a modul SERVIDOR (6/7/96)
 *              Converting all commemts to English  (9/7/96)  
 *		Program transformed to a daemon (17/7/96)	
 *  
 *    Ready for first alpha distribution (30/7/96)
 */      


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <signal.h>
#include <syslog.h>

#include <sys/types.h>
#include <sys/param.h>

#include <linux/timer.h>
#include <linux/socket.h>
#include <linux/sockios.h>
#include <linux/in.h>
#include <linux/inet.h>
#include <linux/netdevice.h> 
#include <linux/ip.h>
#include <linux/icmp.h>
#include <linux/tcp.h>
#include <linux/udp.h>
#include <linux/skbuff.h>
#include <linux/firewall.h>
#include <linux/ip_fw.h>

#include "masq.h"

__u8 rcvbuf[SRV_RCVBUFSIZE];            /* Send buffer */
__u8 sndbuf[SRV_SNDBUFSIZE];            /* Receive buffer */



/*
 *      Signal SIGTERM handler
 */

void ending(int sig)
{
	int fd;
	syslog(LOG_INFO,"masqd terminated.\0");
	closelog();
	for (fd=0; fd<NOFILE; fd++)	close (fd);
	exit(0);
}

/*
 *     Logs to syslog facility and ends the program
 */

void logend (int reason)
{
	switch (reason) {
		case SOCKET_ERR: syslog(LOG_ERR, "Unable to open a socket.\0");

				break;
		case CHILD_ERR:  syslog(LOG_ERR, "Unable to create a child process.\0");
				break;
		default:	 syslog(LOG_ERR, "Unknown error ocurred!!\0");
	}
	ending(SIGTERM);
}

/*
 *   SERVER: Opens /proc/net files to gather firewall information 
 */

FILE *open_file(__u16 firewall)
{
	FILE *fitx;    

	switch (firewall) {

		case IP_FW_IN:	if((fitx=fopen(IP_FW_IN_FILE,"r"))==NULL) return NULL;
				break;
		case IP_FW_OUT:	if((fitx=fopen(IP_FW_OUT_FILE,"r"))==NULL) return NULL;
				break;
		case IP_FW_FWD:	if((fitx=fopen(IP_FW_FWD_FILE,"r"))==NULL) return NULL;
				break;
		case IP_FW_MSQ:	if((fitx=fopen(IP_FW_MSQ_FILE,"r"))==NULL) return NULL;
				break;
		default: return NULL;   /* No hauria d'arribar-hi mai */
		}
	return (fitx);
}

/*
 *  SERVER: Proceses info lines from /proc/net files and convert data to
 *          struct ip_fw format to be re-processed or sent to the client
 *
 *          Rule space MUST be allocated by the caller
 */

struct ip_fw *get_rule(FILE *fp, struct ip_fw *rule) 
{

	int i,num;

	num=fscanf(fp, "%lX/%lX->%lX/%lX %16s %lX %hX %hu %hu %lu %lu",                                
		&(rule->fw_src.s_addr), &(rule->fw_smsk.s_addr),
		&(rule->fw_dst.s_addr), &(rule->fw_dmsk.s_addr), 
		rule->fw_vianame, &(rule->fw_via.s_addr),
		&rule->fw_flg, &rule->fw_nsp, &rule->fw_ndp,
		&rule->fw_pcnt, &rule->fw_bcnt);
	for (i = 0; i < IP_FW_MAX_PORTS; i++)
	num+=fscanf(fp, "%hu", &rule->fw_pts[i]);
	num+=fscanf(fp, " A%hX X%hX", &rule->fw_tosand, &rule->fw_tosxor);
	if (num != RULE_ELEMENTS) return NULL;
	if (!strcmp("-", rule->fw_vianame))
		(rule->fw_vianame)[0] = '\0';
	rule->fw_src.s_addr=htonl(rule->fw_src.s_addr);
	rule->fw_smsk.s_addr=htonl(rule->fw_smsk.s_addr);
	rule->fw_dst.s_addr=htonl(rule->fw_dst.s_addr);
	rule->fw_dmsk.s_addr=htonl(rule->fw_dmsk.s_addr);
	rule->fw_next=NULL;
	return rule;	

}

/*
 *  SERVER: Deletes rule from firewall
 *
 *      Receives a rule number from client and read the nth-rule from /proc/net
 *      files. Then removes it from firewall.
 */

struct ip_fw *del_rule(struct ip_fw *rule, int firewall, int numrule)
{
	int i;
	FILE *fitx;
	char *line;
 
	if ((fitx=open_file(firewall))==NULL) return NULL;
	if ((line=(char *)malloc(LONGLINE+1))==NULL) return NULL; 
	fgets(line,LONGLINE,fitx);
	free (line);

	for(i=0;i<numrule;i++) {
		if (get_rule(fitx,rule)==NULL) return NULL; 
		}
    
	fclose(fitx);
	return(rule);
}

/*
 *   Funcio de punt d'entrada al kernel.
 *   Executa l'acci del programa, en funcio dels parametres.
 */

/*
 *  SERVER:  Server Core! 
 *           Interacts with kernel through setsockopts call
 */ 

int ip_fw_opt(int command, int firewall, void *rule)
{
	int s,err;  
	void *optval;
	int domain, type, protocol;         
	int optlen,level,optname;

                 /* Obrim un socket amb aquests parametres */ 

	domain=AF_INET;
	type=SOCK_RAW;      
	protocol=IPPROTO_RAW; 

	if ((s=socket(domain, type, protocol))<0) return 0; 

	level=SOL_IP;                                  /* Ataquem el nivell IP */
	optname=command | (firewall << IP_FW_SHIFT);   /* Comanda (XXX) */
	optval=rule;	                               /* Parametre de la comanda (Valor) */
         		                               /* Tamany del parametre */
	if (command==IP_FW_POLICY) optlen=sizeof(*(int *)rule); 
		else optlen=sizeof(*(struct ip_fw *)rule);
	if ((err=setsockopt(s, level, optname, optval, optlen))<0) 
		return 0;
   
	close(s);
	return 1;
}

/*
 *  Creates a message to be sent
 */

struct masqmsg *make_msg(int command, int firewall, int policy, struct ip_fw *rule)
{
        struct masqmsg *msg;

        msg=(struct masqmsg *)malloc(sizeof(struct masqmsg));
        msg->command=command;
	msg->firewall=firewall;
	msg->policy=policy;

        switch (command) {

		case SENDLIST:
		case SENDMSQINF:	msg->numrule=1; break;
		default:		msg->numrule=0;
        }

        if (rule) msg->info.rule=rule; 
		else {
			msg->info.rule=NULL;
			msg->numrule=0;
		}

        return (msg);

}

/*
 *  Sends a message!
 */

void send_msqmsg(int sk, struct masqmsg *msg)
{
        int i=0;
	__u32 crc;
        sndbuf[i++]=msg->command;
        sndbuf[i++]=msg->firewall;
        sndbuf[i++]=msg->policy>>8;
        sndbuf[i]=msg->policy&0x0F;
	i+=2;
        sndbuf[i++]=msg->flag;
        sndbuf[i++]=msg->numrule;
        if (msg->numrule && msg->command!=SENDMSQINF)
		i+=datacpy(&sndbuf[i],msg->info.rule,sizeof(struct ip_fw));
        if (msg->command==SENDMSQINF)
		i+=datacpy(&sndbuf[i],msg->info.masq,sizeof(struct masqinf));

        sndbuf[4]=i+sizeof(__u32);
        crc=checkcrc(&sndbuf,i);
	i+=datacpy(&sndbuf[i],&crc,sizeof(__u32));
 
        send(sk,&sndbuf, i, 0);
	free (msg);
}

/*
 *   Send a confirm message (SRVACK)
 */

inline void send_ack(int sk)
{
	send_msqmsg(sk,make_msg(SRVACK,0,0,NULL)); 
}

/*
 *   Send a Non conform message (SRVNAK)
 */

inline void send_nak(int sk)
{
	send_msqmsg(sk,make_msg(SRVNAK,0,0,NULL));
} 

/*
 *  Host authentification. It should contro if that client is allowed to conect
 *
 *  TODO: Implement the host authentification, comparing the name retrieved 
 *        by the DNS with a local file 
 *
 */

int allow_client(__u32 claddr)
{
	struct hostent *host;
	char *straddr;

	straddr=(char *)malloc(BUFSIZE);	
	straddr[3]=claddr>>24;
	straddr[2]=(claddr<<8)>>24;
	straddr[1]=(claddr<<16)>>24;
	straddr[0]=(claddr<<24)>>24;
#ifdef SECURE_IP
	host=(struct hostent *)gethostbyaddr(straddr,4,AF_INET);
	free(straddr);

	if (!host) return 0; 
#else
	free(straddr);
#endif 
	return 1; 
}

inline void reject_client(int sk)
{
	send_msqmsg(sk,make_msg(REJECT,0,0,NULL));		
}


/* 
 *  Checks username & password on the local users file
 */

int verify_user(char *user, char *pw)
{
	FILE *fd=fopen(USER_FILE,"r");
	int i,j;
	char *fuser, *fpw, *ret;
	char *line=(char *)malloc(80);
	
	fuser=(char *)malloc(USER_LONG);
	ret=fgets(line,80,fd);
	while (ret!=NULL) {
		i=0;
		while (line[i]!=':') fuser[i]=line[i++];
		fuser[i++]='\0';
		if (!(strcmp(user,fuser))) break;
		ret=fgets(line,80,fd);
	}
	free (fuser);
	close (fd);	
	if (ret==NULL) {   /*User does not exist */
		free (line);
		return 0;	
	}
	j=0;
	fpw=(char *)malloc(128);
	while (line[i]!=':') fpw[j++]=line[i++];
	fpw[j]='\0';
	if (strcmp(pw,fpw)) { /* Incorrect password */
		free (line);
		return 0;
	}
	free (fpw);
	free (line);
	return 1;	/* Identification succesfull */
}
	

/*
 *  Requests user identify to the client
 */

int request_identify(int sk)
{
	int nrec;
	char *info, *userrcv, *pwcryp;
	struct masqmsg *rcvmsg;

	send_msqmsg(sk,make_msg(IDENTIFY,0,0,NULL));	
	rcvmsg=(struct masqmsg *)malloc(sizeof(struct masqmsg));
	if (!(recv_msqmsg(sk,rcvmsg))) {
		free (rcvmsg);
		return 0;	
	}
        if (!(rcvmsg->command==IDENTITY)) {
		send_nak(sk);
		free (rcvmsg);
		return 0;
	}
 	if (rcvmsg->numrule==1) info=rcvmsg->info.raw;
		else {
			send_nak(sk);
			free (rcvmsg);
			return 0;
		}
	free (rcvmsg); 
	userrcv=(char *)malloc(USER_LONG);
	pwcryp=(char *)malloc(128);
	sscanf(info,"%s\n%s",userrcv,pwcryp);	
	if (!verify_user(userrcv, pwcryp)) return 0;
	send_msqmsg(sk,make_msg(ACCEPT,0,0,NULL));	
	return 1;
}

/*
 *   Receives a message from the client
 *   msg MUST be allocated!
 */


int recv_msqmsg(int sk, struct masqmsg *msg)
{
	int nrec,num=0;
	char *info;
	
        nrec=recv(sk, &rcvbuf, SRV_RCVBUFSIZE, 0);
        if (nrec<MINPKTSIZE) return 0;

	msg->command=rcvbuf[0];
	msg->firewall=rcvbuf[1];
	msg->policy=rcvbuf[2]<<8|rcvbuf[3];
	msg->longpkt=rcvbuf[4];
	msg->flag=rcvbuf[5];
	msg->numrule=rcvbuf[6];
	if (msg->numrule && msg->command!=DELRULE) {
		info=(char *)malloc(SRV_RCVBUFSIZE); 
		num=datacpy(info,&rcvbuf[7],msg->longpkt-MINPKTSIZE);
	}
	else 
		info=NULL;
	msg->info.raw=info;
	datacpy(&(msg->crc),&rcvbuf[7+num],sizeof(__u32));
        if (msg->crc==checkcrc(rcvbuf,7+num)) return 1;
                else return 0;
}

/*
 *    Performs the 'check before act'
 */

int check(int sk, int firewall, int numrule, struct ip_fw *rule)
{
	struct masqmsg *msg;
	int nrec;

	msg=make_msg(CONFIRMOP,firewall,0,rule);
	msg->numrule=numrule;
	send_msqmsg(sk,msg);
       	nrec=recv(sk, &rcvbuf, SRV_RCVBUFSIZE, 0);
       	if (nrec>=MINPKTSIZE) msg=(struct masqmsg *)&rcvbuf;
        	else return 0;
	if (!(msg->command==CLNACK)) return 0;
		else return 1;
}

/*
 *   Performs the ADDRULE operation
 */

int do_addrule(int sk, int firewall, struct ip_fw *rule)
{
	if (!check(sk,firewall,1,rule)) return 0;

	if (ip_fw_opt (IP_FW_APPEND,firewall,(void *)rule)>0) { 
		send_ack(sk);
		return 1;
	}
	else {
		send_nak(sk);
		return 0;
	}
}

/*
 *   Performs the DELRULE operation
 */

int do_delrule(int sk, int firewall, int numrule, struct ip_fw *rule)
{

	if (del_rule(rule, firewall, numrule)==NULL) {
		send_nak(sk);
		return 0;
	}	
	if (!check(sk,firewall,numrule,rule)) return 0;
	
	if (ip_fw_opt (IP_FW_DELETE,firewall,(void *)rule)>0) { 
		send_ack(sk);
		return 1;
	}
	else {
		send_nak(sk);
		return 0;
	}
}

/*
 *   Fetches rule lines from /proc/net files and send them to the client
 */   

int send_rules(int sk, int firewall, struct ip_fw *rule)
{
	int i,nrec,tries,oneline=TRUE;
	FILE *fitx;
	char *line,c;
	struct masqmsg *rcvmsg;
	int policy;
            
	if ((fitx=open_file(firewall))==NULL) {
		send_msqmsg(sk,make_msg(SRVNAK,0,0,NULL));
		return 0;
	}
	if ((line=(char *)malloc(LONGLINE+1))==NULL) return 0;

	while (!(isdigit( (c=fgetc(fitx)) )));
	ungetc(c,fitx);
	fgets(line,LONGLINE,fitx); 
	sscanf(line,"%d",&policy);
	free (line);
      
	while (ungetc(fgetc(fitx),fitx)!=EOF) {
		oneline=FALSE;
		if (get_rule(fitx,rule)==NULL) break;
		tries=NUM_RETRY;
		do {
			tries--;
			send_msqmsg(sk,make_msg(SENDLIST,firewall,policy,rule));
        		nrec=recv(sk, &rcvbuf, SRV_RCVBUFSIZE, 0);
        		if (nrec>=MINPKTSIZE) rcvmsg=(struct masqmsg *)&rcvbuf;
                		else return 0;
		}
        	while (tries && rcvmsg->command==CLNNAK);
 		if (rcvmsg->command!=CLNACK) return 0;
	}       
	if (oneline) {
		send_msqmsg(sk,make_msg(SENDLIST,firewall,policy,NULL));
        	nrec=recv(sk, &rcvbuf, SRV_RCVBUFSIZE, 0);
        	if (nrec>=MINPKTSIZE) rcvmsg=(struct masqmsg *)&rcvbuf;
               		else return 0;
		if (!(rcvmsg->command==CLNNAK)) return 0;
	}		
	fclose(fitx);   
	send_msqmsg(sk,make_msg(SRVACK,0,0,NULL));
	return 1;
}


/*
 *   Fetches masquerade lines from /proc/net/ip_masquerade file and send 
 *	them to the client
 */   

int send_msqinf(int sk, int firewall)
{
	int i,nrec,tries,oneline=TRUE;
	FILE *fitx;
	char *line,*strprot;
	__u32 saddr, daddr;
	__u16 *sport, *dport, *mport, k; 
	struct masqmsg *rcvmsg;
	struct masqinf *mq;
            
	if ((fitx=open_file(firewall))==NULL) {
		send_msqmsg(sk,make_msg(SRVNAK,0,0,NULL));
		return 0;
	}
	if ((mq=(struct masqinf *)malloc(sizeof(struct masqinf)))==NULL) 
		return 0;
	if ( (line=(char *)malloc(LONGLINE+1) )==NULL) return 0;
	if ( (strprot=(char *)malloc(LONGLINE+1) )==NULL) return 0; 
	if ( (sport=(unsigned short *)malloc(sizeof(__u16)) )==NULL) return 0;
	if ( (dport=(unsigned short *)malloc(sizeof(__u16)) )==NULL) return 0;
	if ( (mport=(unsigned short *)malloc(sizeof(__u16)) )==NULL) return 0;


	fgets(line,LONGLINE,fitx);
	while (ungetc(fgetc(fitx),fitx)!=EOF) {
		i=-1;
		fgets(line,LONGLINE,fitx);
		while(line[++i]==' ');
		if (isalpha(line[i])) {
			sscanf(line,"%3s %lX:%hX %lX:%hX %hX %lX %hu %hu %lu", 
				strprot, &saddr, sport,
				&daddr, dport, mport,
				&(mq->initseq), &(mq->delta), &(mq->pdelta), &(mq->exptime));               
			mq->protocol=htons(get_protocol(strprot));
			mq->saddr=htonl(saddr);
			mq->daddr=htonl(daddr);
			mq->sport=htons(*sport);
			mq->dport=htons(*dport);
			mq->mport=htons(*mport);
			send_msqmsg(sk, make_msg(SENDMSQINF,IP_FW_MSQ,0,(struct ip_fw *)mq));
       		 	nrec=recv(sk, &rcvbuf, SRV_RCVBUFSIZE, 0);
       		 	if (nrec>=MINPKTSIZE) rcvmsg=(struct masqmsg *)&rcvbuf;
       	        		else return 0;
			if (!(rcvmsg->command==CLNACK)) return 0;
		}
	}
	fclose(fitx);   
	free (sport); free (dport); free (mport); 
	free (strprot); free (line); free(mq);
	send_msqmsg(sk,make_msg(SRVACK,0,0,NULL));
	return 1;
}		


/*
 *   Performs the RESET operation
 */

int do_reset(int sk,int firewall,struct ip_fw *rule)
{
				
	if (ip_fw_opt (IP_FW_FLUSH,firewall,(void *)rule)>0) { 
		send_ack(sk);
		return 1;
	}
	else {
		send_nak(sk);
		return 0;
	}
}

/*
 *   Performs the SETPOLICY operation
 */

int do_setpolicy(int sk, int firewall,int policy)
{
				
	if (ip_fw_opt (IP_FW_POLICY,firewall,(void *)&policy)>0) { 
		send_ack(sk);
		return 1;
	}
	else {
		send_nak(sk);
		return 0;
	}
}

/*
 *   Performs the INITIALIZE operation
 */

int do_initialize(int sk)
{
	int policy=IP_FW_F_ACCEPT;
				
	 if(ip_fw_opt (IP_FW_POLICY,IP_FW_IN,(void *)&policy)>0 && 
	    ip_fw_opt (IP_FW_POLICY,IP_FW_OUT,(void *)&policy)>0 && 
	    ip_fw_opt (IP_FW_POLICY,IP_FW_FWD,(void *)&policy)>0 && 
	    ip_fw_opt (IP_FW_FLUSH,IP_FW_IN,(void *)&policy)>0 &&
	    ip_fw_opt (IP_FW_FLUSH,IP_FW_OUT,(void *)&policy)>0 &&
	    ip_fw_opt (IP_FW_FLUSH,IP_FW_FWD,(void *)&policy)>0 ) {
	    	send_ack(sk);
	    	return 1;
	}
	else {
		send_nak(sk);
		return 0;
	}
}

/*
 *   Performs the GENSCRIPT operation
 */

int do_genscript(int sk)
{
	char *command=malloc(sizeof(char)*BUFSIZE);
	strcpy(command,BINDIR);
	strcat(command,"/list2script\0");
	if (!(system(command))) {
		free (command);
		send_ack(sk);
		return 1;
	}
	else {	
		free (command);
		send_nak(sk);
		return 0;
	}
}


/*
 *  Processes a new incommed message 
 */

void proces_msqmsg(int sk, struct masqmsg *msg)
{
	int ok;
	int command=msg->command;
	int firewall=msg->firewall;
	int policy=msg->policy;
	int numrule=msg->numrule;
	struct ip_fw *rule;

	if (msg->info.rule) rule=msg->info.rule;
		else rule=(struct ip_fw *)malloc(sizeof(struct ip_fw));
 	switch (command) {
		
		case ADDRULE:		ok=do_addrule(sk,firewall,rule);
					break;
		case DELRULE:		ok=do_delrule(sk,firewall,numrule,rule);
					break;
		case GETLIST:		ok=send_rules(sk,firewall,rule);
					break;
		case GETMSQINF:		ok=send_msqinf(sk,firewall);
					break;
		case RESET:		ok=do_reset(sk,firewall,rule);
					break;
		case SETPOLICY:		ok=do_setpolicy(sk,firewall,policy);
					break;
		case INITIALIZE:	ok=do_initialize(sk);
					break;
		case DISCONNECT:	send_ack(sk);
					close(sk);
					break;
		case GENSCRIPT:		do_genscript(sk);
					break;
	}			
	free (rule);	
	if (!ok) close(sk);
}

int Child_server_main(int sk, struct sockaddr_in client)
{
	char buffer[100];
	struct masqmsg *msg;
	int idcount=LOGIN_COUNT;
	
	/*
	 * Checks the IP address of the client to see if it is allowed
	 */

	if (allow_client(client.sin_addr.s_addr)!=TRUE) {
		reject_client(sk);
		close(sk);
		exit(-1);
	}

	/*
	 * Performs an identity check with ID & password
	 */

	while (idcount && request_identify(sk)!=TRUE) idcount--;
	if (!idcount) {
		reject_client(sk);
		close(sk);
		exit(-1);
	}

	msg=(struct masqmsg *)malloc(sizeof(struct masqmsg));
	while ((recv_msqmsg(sk,msg))) proces_msqmsg(sk,msg);
	free (msg);
	exit(0);
        
}

/*
 *  Routine to transfor this program into a daemon
 */

int become_daemon(int ignsigcld)
{
	register int childpid,fd;
	
	signal(SIGTTOU, SIG_IGN);
	signal(SIGTTIN, SIG_IGN);
	signal(SIGTSTP, SIG_IGN);	
	if ((childpid=fork())<0)
		logend(CHILD_ERR);
	else if (childpid>0)  { /* parent */
		closelog();
		exit(0);
	}

	/* Child process */

	/* Re starting logging services */
	closelog();
	openlog("masqd:\0",LOG_CONS | LOG_PID, LOG_DAEMON);	
	if (setpgrp()==-1)
		logend(ERROR);
	signal(SIGHUP, SIG_IGN);
	if ((childpid=fork())<0)
		logend(CHILD_ERR);
	else if (childpid>0) { /* old child new parent */
		closelog();
		exit(0);
	}

	/* New child process */

	/* Close old log services */
	closelog();	

	/* Close any open file descriptors */
	for (fd=0; fd<NOFILE; fd++)	close (fd);

	/* Starting logging again */
	openlog("masqd:\0",LOG_CONS | LOG_PID, LOG_DAEMON);	

	/* Change current directory to root */
 	chdir("/");

	/* Clear any file mode creation mask */
	umask(0);

	if (ignsigcld) signal(SIGCLD, SIG_IGN);	
	
}


int main ()
{
	int ignsigcld=1;
	int sk,sk2,ncon,ser_addr_len, cli_addr_len,pid;
        struct sockaddr_in ser_addr, cli_addr;
	
	setuid(0);
	if(geteuid() != 0) {
		printf("This program requires root privileges\n");
		exit(0);
	}

	openlog("masqd:\0",LOG_CONS | LOG_PID, LOG_DAEMON);	
	syslog(LOG_INFO,"masqd started.\0");
	signal (SIGTERM, ending);
	become_daemon(ignsigcld);  

	if ( (sk = socket (AF_INET, SOCK_STREAM, IPPROTO_TCP)) == -1 ) 
		logend (SOCKET_ERR);
                
	ser_addr.sin_family = AF_INET;
        ser_addr.sin_port=ntohs(SRVPORT);
        ser_addr.sin_addr.s_addr=inet_addr("0.0.0.0\0");
        ser_addr_len = sizeof(ser_addr);


   	if ( bind (sk, (struct sockaddr *)&ser_addr, ser_addr_len) == -1 )
		logend (SOCKET_ERR);
        
        listen (sk, 1);

	for (;;) {
		ncon++;
		cli_addr_len = sizeof (cli_addr);
 
		if ( (sk2 = accept (sk,(struct sockaddr *)&cli_addr, &cli_addr_len)) == -1 )
			logend (SOCKET_ERR);
		if ( (pid = fork()) == -1 )
			logend (CHILD_ERR);
		else {
			if ( pid == 0 )   
				Child_server_main(sk2,cli_addr);
			else waitpid(pid,NULL,0);
		}  
	}
}

