/*
 *    Tercera utilitat de projecte sobre Conversio d'adreces IP
 *    
 *    Tipus de Programa: CLIENT TEXTE
 *    Objectiu: Manteniment (afegir, esborrar, consultar) de "rules" del Firewall 
 *    Data inici: 27/5/96
 * 
 *    Fase2: Separaci modular CLIENT/SERVIDOR (4/7/96)
 *
 *    27/10/97: Added support for bidirectional rules & range of ports 
 */      


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <linux/types.h>
#include <linux/timer.h>
#include <linux/socket.h>
#include <linux/sockios.h>
#include <linux/in.h>
#include <linux/inet.h>
#include <linux/netdevice.h> 
#include <linux/ip.h>
#include <linux/icmp.h>
#include <linux/tcp.h>
#include <linux/udp.h>
#include <linux/skbuff.h>
#include <linux/firewall.h>
#include <linux/ip_fw.h>

#include "masq.h"

/*
 *   CLIENT: Explain syntax to the user 
 */

void sintaxi(void)
{
   printf("
  The sintax:
 
   # [l]masq [ type ] command parameters [ -hostname ]
   
     where:         
             type:        -i | -o | -f | -m  (it is -f if not specified) 
             commmad:     add | del | policy | list | reset | initialize 
             parameters:  depending on command:   
                     add:     [ protocol ] [-b] 
                              from-addr netmask  [ from-port | 0 | iport-eport]
                              to-addr   netmask  [ to-port   | 0 | iport-eport]
                              action
                     del:     rule_number 
		     policy:  action
                     list:       <no parameters>
                     reset:      <no parameters>
                     initialize: <no parameters>
		     genscript:  <no parameters, only masq>
 
             	     protocol:   tcp | udp | icmp
             	     action:     accept | convert | reject | block  
                                     
                                                                    \n");
}

/*      
 * CLIENT/SERVER: Error handling
 */

terminate(int nerror)
{
	switch (nerror)
	{
		case SILENT:	   break;		   
		case PARAMS_ERR:   printf("Incorrect parameters.\n\0");
                                   sintaxi();  
                                   break;
		case SOCKET_ERR:   printf("Unable to open a socket.\n\0");
                                   break;
		case SOCKOPT_ERR:  printf("Incorrect kernel call.\n\0");
                                   break;
		case MEM_ERR:      printf("Insuficient memory.\n\0");
                                   break;
                case FILE_ERR:     printf("Masquerade or proc filesystem not active.\n\0");
                                   break;
                case CHILD_ERR:    printf("Unable to create a child process.\n\0");
				   break;
                case HOST_ERR:     printf("Unable to connect to host.\n\0");
				   break;
                case IDENT_ERR:    printf("Server does not identify you.\n\0");
				   break;
                case PROTO_ERR:    printf("Protocol not followed!.\n\0");
				   break;
		case RANGE_ERR:	   printf("In range ports, end port must be greater than initial port!\0");
				   break;		
                default:           printf("Unknown error.\n");
	};
	exit(ERROR);
}	


/*
 *  Funcio que retorna el codi d'accio a realitzar a partir d'un string	
 * 
 *  Els valors possibles per a action son (veure firewall.h):
 *    acceptar     IP_FW_F_ACCEPT                     0x4  -> Inicial Linux
 *    convertir    IP_FW_F_MASQ + IP_FW_F_ACCEPT    0x204 
 *    expulsar     IP_FW_F_ICMPRPL                  0x100
 *    bloquejar                                         0       
 */		 

/*
 *  CLIENT: Syntax parse function
 */

/*
 *   Returns an action value depending on the string stract.
 */

unsigned short get_action(char *stract)
{
      unsigned short action,i; 
	
      for (i=0;i<strlen(stract);i++) stract[i]=tolower(stract[i]);
   
      if (!(strcmp(stract,"accept\0"))) action=IP_FW_F_ACCEPT;
        else if (!(strcmp(stract,"convert\0"))) action=IP_FW_F_MASQ|IP_FW_F_ACCEPT;
          else if (!(strcmp(stract,"reject\0"))) action=IP_FW_F_ICMPRPL;
            else if (!(strcmp(stract,"block\0"))) action=0;
              else terminate(PARAMS_ERR);  
      return(action);
}   

/*
 *  Funcio que retorna el string associat al codi d'accio d'entrada
 *
 *  PRE: Requereix vector amb espai per a 16 chars.
 */ 

/* 
 *   CLIENT: Numeric to readable syntax convert
 */ 

/*   
 *  Returns an action string depending on the parameter action
 *  
 *  MUST have allocated space at stract (let it 16 chars)
 */

char *set_action(char *stract, unsigned short action)
{
      int i;	

      if ((action&(IP_FW_F_MASQ+IP_FW_F_ACCEPT))==IP_FW_F_MASQ+IP_FW_F_ACCEPT) strcpy(stract,"convert\0");
	else if ((action&IP_FW_F_ACCEPT)==IP_FW_F_ACCEPT) strcpy(stract,"accept\0");
          else if ((action&IP_FW_F_ICMPRPL)==IP_FW_F_ICMPRPL) strcpy(stract,"reject\0");
            else strcpy(stract,"block\0");
      return(stract);
}


/*
 *   Retorna un valor en funcio del parametre protocol
 */
 
/*
 *   CLIENT: Syntax parse function: Readable form to numeric form.
 */

/*
 *   Returns a protocol value depending on the string strprot.
 */
 
unsigned short get_protocol(char *strprot)
{
      unsigned short proto,i; 
	
      for (i=0;i<strlen(strprot);i++) strprot[i]=tolower(strprot[i]);

      if (!(strcmp(strprot,"tcp\0"))) proto=IP_FW_F_TCP;
        else if (!(strcmp(strprot,"udp\0"))) proto=IP_FW_F_UDP;
          else if (!(strcmp(strprot,"icmp\0"))) proto=IP_FW_F_ICMP;
      
	return (proto);
}   


/*
 *   Retorna un string de protocol en funcio del valor entrada
 *
 *  PRE: Requereix vector amb espai per a 8 chars.
 */

 /*
  *   CLIENT: Numeric to readable syntax convert
  */


/*   
 *  Returns a protocol string depending on the parameter proto
 *  
 *  MUST have allocated space at strprot (minimum 5 chars)
 */

char *set_protocol(char *strprot, unsigned short proto)
{
      unsigned short i; 
	
      if (proto==IP_FW_F_TCP) strcpy(strprot,"tcp\0");
        else if (proto==IP_FW_F_UDP) strcpy(strprot,"udp\0");
          else if (proto==IP_FW_F_ICMP) strcpy(strprot,"icmp\0");
              else  strcpy(strprot,"all\0");
      return(strprot);
}   

/*
 *   Returns a command value depending on the string strcmd.
 */

unsigned short get_command(char *strcmd, unsigned int length)
{ 
	char *param2;
	unsigned int command,i;

	if ((param2=(char *)malloc(1+length))==NULL) terminate (MEM_ERR);
   	for (i=0;i<length;i++) param2[i]=tolower(strcmd[i]);
   	if (!(strcmp(param2,"add\0"))) command=ADDRULE;
		else if (!(strcmp(param2,"del\0"))) command=DELRULE;
			else if (!(strcmp(param2,"policy\0"))) command=SETPOLICY;
				else if (!(strcmp(param2,"list\0"))) command=GETLIST;
					else if (!(strcmp(param2,"reset\0"))) command=RESET;
						else if (!(strcmp(param2,"initialize\0"))) command=INITIALIZE;
							else if (!(strcmp(param2,"genscript\0"))) command=GENSCRIPT;
								else terminate(PARAMS_ERR);
	free (param2);
	return(command);	
}

/*   
 *  Returns a firewall name string depending on the parameter firewall 
 *  
 *  MUST have allocated space at strfire  (let it 16 chars)
 */

char *set_firewall(char *strfire, unsigned short firewall)
{
      int i;	

      if (firewall==IP_FW_IN) strcpy(strfire,"input\0");
        else if (firewall==IP_FW_OUT) strcpy(strfire,"output\0");
          else if (firewall==IP_FW_FWD) strcpy(strfire,"forward\0");
              else  {free (strfire); strfire=NULL; } 
      return(strfire);
}

char set_fw(unsigned short firewall)
{
      int i;	

      if (firewall==IP_FW_IN) return 'i'; 
        else if (firewall==IP_FW_OUT) return 'o';
          else if (firewall==IP_FW_FWD) return 'f';
		else return 0;
}

/*

/*
 * Function to analize and extract ports from command line
 */

unsigned short extract_ports(char *strports, struct ip_fw *rule, unsigned short nport)
{ 
	unsigned int iport, eport;
	char *strport2;
	if ((iport=atoi(strports))==0)  return 0;
	if ((strport2=strchr(strports,'-'))==NULL) {
		rule->fw_pts[nport]=iport;
		return 1;
	} else {
		strport2[0]='0';
		eport=atoi(strport2);
		if (iport>eport) terminate (RANGE_ERR);
		rule->fw_pts[nport]=iport;
		rule->fw_pts[nport+1]=eport;
		return 2;
	}
}
		
/*
 * Function to show ranges of ports in list command
 *
 * REQ: strport must be alocated with at least 12 bytes "00000-00000\0"
 */

char *list_ports(char *strport, struct ip_fw *rule, int nport, int indexport, int padd,int source)
{
	int i;
	char *strp=malloc(sizeof(char)*10);
	switch (nport) {
	case 0: 	strcpy(strport,"ALL         \0");
			break;
	case 1:		sprintf(strport,"%d",rule->fw_pts[indexport]);
			if (padd) for (i=12-strlen(strport);i>0; i--) strcat(strport," \0");	
			break;
	case 2:		sprintf(strp,"%d",rule->fw_pts[indexport]);
			strcpy(strport,strp);
			if (((source) && (rule->fw_flg&IP_FW_F_SRNG)) || (!(source) && (rule->fw_flg&IP_FW_F_DRNG))) strcat(strport,"-\0");
			else strcat(strport,",\0");
			sprintf(strp,"%d",rule->fw_pts[indexport+1]);
			strcat(strport,strp);
			if (padd) for (i=12-strlen(strport);i>0; i--) strcat(strport," \0");	
			break;
	}
	free (strp);
	return strport;
}

/*
 *  Copies size bytes of one existing buffer to another;
 *  both buffers MUST be allocated	
 */

int datacpy (char *dst, char *src, int size)
{
	int i;
	for (i=0;i<size;i++) ((char *)dst)[i]=((char *)src)[i];
	return i;
} 

/*
 *  Calculates the CRC of one byte
 */

unsigned int updcrc(unsigned int crc, int c, unsigned int mask)
{
        int i;
        c<<=8;
        for(i=0;i<8;i++)
        {
                if((crc ^ c) & 0x8000) crc=(crc<<1)^mask;
                else crc<<=1;
                c<<=1;
        }
        return crc;
}

/*
 *   Calculates the CRC of one buffer of bytes 
 */

#define MTT	0x1021		/* crc-ccitt mask */

__u32 checkcrc(char *buffer, int longbuf)
{
	int i;
	__u32 crc=0;

	for (i=0; i<longbuf; i++) 
                crc=updcrc(crc,buffer[i],MTT);
	return crc;
}
